from dune.perftool.pdelab.argument import pymbolic_coefficient
from dune.perftool.generation import (domain,
                                      get_counter,
                                      iname,
                                      instruction,
                                      silenced_warning,
                                      )
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.pdelab.spaces import name_lfs
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )

from loopy import Reduction

from pytools import product


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


# TODO this code is WIP and mainly used for experiments.
def start_sumfactorization(element, container, restriction, component):
    from dune.perftool.sumfact.amatrix import (AMatrix,
                                               quadrature_points_per_direction,
                                               basis_functions_per_direction,
                                               name_theta,
                                               )

    theta = name_theta()
    m = quadrature_points_per_direction()
    n = basis_functions_per_direction()
    a_matrix = AMatrix(theta, m, n)
    a_matrices = (a_matrix, a_matrix)

    initialize_buffer("buffer",
                      base_storage_size=product(max(mat.n, mat.m) for mat in a_matrices),
                      num=2
                      )


    number_basis = product(mat.n for mat in a_matrices)
    shape = (n,)
    inp = get_buffer_temporary("buffer",
                               shape=shape)
    silenced_warning('read_no_write({})'.format(inp))

    # Write initial coefficients into buffer
    basisiname = sumfact_iname(number_basis, "basis")
    lfs = name_lfs(element, restriction, component)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),))
    from dune.perftool.loopy.stages import stage_insn
    insn_dep = instruction(assignee = assignee,
                           expression = coeff,
                           depends_on = frozenset({stage_insn(0)}),
                           )

    return sum_factorization_kernel(a_matrices, inp, "buffer", insn_dep)


def sum_factorization_kernel(a_matrices, inp, buffer, insn_dep):
    """
    Calculate a sum factorization matrix product.

    Y = A_{d-1}*...*A_0*X

    where X is the input matrix and Y is the output variable.

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    inp: A temporary that contains the input matrix
    buffer: A string identifying the flip flop buffer in use
        for intermediate results.
    """
    for l, a_matrix in enumerate(a_matrices):
        # Get a temporary that interprets the base storage of the input
        # as a column-major matrix. In later iteration of the amatrix loop
        # this reinterprets the output of the previous iteration.
        inp_shape = (a_matrix.n, product(mat.m for mat in a_matrices[:l]) * product(mat.n for mat in a_matrices[l + 1:]))
        inp = get_buffer_temporary(buffer,
                                   shape=inp_shape,
                                   dim_tags="f,f")

        # The input temporary will only be read from, so we need to silence the loopy warning
        silenced_warning('read_no_write({})'.format(inp))

        switch_base_storage(buffer)

        # Get a temporary that interprets the base storage of the output
        # as row-major matrix
        out_shape = (a_matrix.m, product(mat.m for mat in a_matrices[:l]) * product(mat.n for mat in a_matrices[l + 1:]))
        out = get_buffer_temporary(buffer,
                                   shape=out_shape,
                                   dim_tags="c,c")

        # Get the inames needed for one matrix-matrix multiplication
        i = sumfact_iname(out_shape[0], "row")
        j = sumfact_iname(out_shape[1], "col")
        k = sumfact_iname(a_matrix.n, "red")

        # Construct the matrix-matrix-multiplication expression a_ik*in_kj
        from dune.perftool.sumfact.amatrix import ColMajorAccess
        prod = Product((Call(ColMajorAccess(a_matrix), (Variable(i), Variable(k))),
                        Subscript(Variable(inp), (Variable(k), Variable(j)))
                        ))

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        insn_dep = instruction(assignee=Subscript(Variable(out), (Variable(i), Variable(j))),
                               expression=Reduction("sum", k, prod),
                               forced_iname_deps=frozenset({i, j}),
                               forced_iname_deps_is_final=True,
                               depends_on=frozenset({insn_dep}),
                               )

    return out
